using System;
using System.IO;
using System.Collections;

namespace AVIInfo
{
	/// <summary>
	/// Summary description for AVIWrapper.
	/// </summary>
	public class AVIWrapper
	{
        private string m_sAVIFile = null;

        private long m_iFileSize;
        private long m_iFileSize32bit;

        private string m_sWarnings;
        private string m_sErrors;

        ArrayList m_lhHeaderArray = new ArrayList();
        ArrayList m_lhStreamArray = new ArrayList();
        ArrayList m_lhMovieArray = new ArrayList();
        ArrayList m_aiArray = new ArrayList();

        public bool m_bDiagnostics = false;
        private int m_iList = 0;

        private int m_iCurrentStreamHeaderType = 0;

        public AVIBasicInfo m_aviBasicInfo = new AVIBasicInfo();
        public AVIStreamHeader m_ashAudio = new AVIStreamHeader();
        public AVIStreamHeader m_ashVideo = new AVIStreamHeader();
        public AVIStreamFormatAudio m_asfAudio = new AVIStreamFormatAudio();
        public AVIStreamFormatVideo m_asfVideo = new AVIStreamFormatVideo();

        public OpenDMLIndex m_odmlIndexVideo = null;
        public OpenDMLIndex m_odmlIndexAudio = null;

        public ArrayList m_avi1Index = null;

        public ArrayList m_odmlFullIndex = null;
        public long m_iMasterODMLFrameIndex = 0;

        private long[] m_KeyFrames = null;

		public AVIWrapper()
		{
		}

        public string Warnings
        {
            get { return m_sWarnings; }
        }

        public string Errors
        {
            get { return m_sErrors; }
        }

        private void Diag(string s)
        {
            if (m_bDiagnostics)
            {
//                string sFormat = string.Format("{{0,{0}:G}}", m_iList);
//                Console.WriteLine("Diag: " + sFormat + s, "");
                Console.WriteLine("Diag: " + s);
            }
        }

        private void PosDiag(string s, long l)
        {
            Diag(string.Format("0x{0:X16} - {1}", l, s));
        }

        private int PushList()
        {
            m_iList += 2;
            if (m_iList > 40)
                m_iList = 40;
            return m_iList;
        }

        private int PopList()
        {
            m_iList -= 2;
            if (m_iList < 0)
                m_iList = 0;
            return m_iList;
        }

        private int NewList()
        {
            m_iList = 0;
            return m_iList;
        }

        public void SetDefaults()
        {
            m_sWarnings = "";
            m_sErrors = "";

            m_sAVIFile = null;

            m_iFileSize = 0;
            m_iFileSize32bit = 0;

            m_lhHeaderArray = new ArrayList();
            m_lhStreamArray = new ArrayList();
            m_lhMovieArray = new ArrayList();
            m_aiArray = new ArrayList();

            m_bDiagnostics = false;
            m_iList = 0;

            m_iCurrentStreamHeaderType = 0;

            AVIBasicInfo m_aviBasicInfo = new AVIBasicInfo();
            AVIStreamHeader m_ashAudio = new AVIStreamHeader();
            AVIStreamHeader m_ashVideo = new AVIStreamHeader();
            AVIStreamFormatAudio m_asfAudio = new AVIStreamFormatAudio();
            AVIStreamFormatVideo m_asfVideo = new AVIStreamFormatVideo();

            m_odmlIndexVideo = null;
            m_odmlIndexAudio = null;

            m_avi1Index = null;

            m_odmlFullIndex = null;
            m_iMasterODMLFrameIndex = 0;

            m_KeyFrames = null;
        }

        public bool SetAVI(string sFile)
        {
            SetDefaults();

            if (!File.Exists(sFile))
                return false;

            FileStream fs;

            m_sAVIFile = sFile;
            try
            {
                fs = File.OpenRead(sFile);
                Diag("Opened file \"" + sFile + "\"");
            }
            catch
            {
                return false;
            }

            m_iFileSize = fs.Length;

            BinaryReader reader = new BinaryReader(fs);

            bool bValid = ReadAVI(reader);

            fs.Close();

            return bValid;
        }

        private bool ReadAVI(BinaryReader reader)
        {
            Diag("Beginning AVI parsing");

            if (!ReadRIFF(reader))
                return false;

            Diag("Read RIFF header");

            while (reader.PeekChar() != -1)
            {
                if (!ReadChunk(reader))
                    return false;
            }

            Diag("EOF reached");

            if (m_odmlIndexVideo != null)
            {
                ReadOpenDMLFullIndex(m_odmlIndexVideo, reader);
            }

            if (m_lhHeaderArray.Count < 1)
            {
                AddError("No header list chunks were found.");
                return false;
            }

            if (m_lhMovieArray.Count < 1)
            {
                AddError("No movie list chunks were found.");
                return false;
            }

            if (m_aiArray.Count < 1)
            {
                AddError("No AVI index chunks were found.");
                return false;
            }

            return true;
        }

        private void SkipChunk(BinaryReader reader, long iSize)
        {
            reader.BaseStream.Position += iSize;
        }

        private bool ReadChunk(BinaryReader reader)
        {
            long iPosition = reader.BaseStream.Position;

            int iType = reader.ReadInt32();
            int iSize = reader.ReadInt32();

            if (iType == StringInt("JUNK"))
            {
                PosDiag("JUNK chunk (ignored)", iPosition);

                SkipChunk(reader, iSize);
                return true;
            }

            if (iType == StringInt("LIST"))
            {
                return ReadList(reader, iSize);
            }
            else if (iType == StringInt("idx1"))
            {
                PosDiag("AVI index", iPosition);

                m_aiArray.Add(ReadAVIIndex(reader, iSize));
                return true;
            }
            else if (iType == StringInt("RIFF"))
            {
                // Another RIFF chunk; probably open DML
                iType = reader.ReadInt32();

                if (iType == StringInt("AVIX"))
                {
                    Diag("Detected OpenDML segment");
                } 
                else
                {
                    AddWarning("Additional RIFF header found, but AVIX chunk missing");
                }
            }
            else
            {
                PosDiag("(unknown)", iPosition);
                AddWarning(string.Format("An unknown chunk type (\"{0}\") was found.", IntString(iType)));
                SkipChunk(reader, iSize);
            }

            return true;
        }

        private bool ReadList(BinaryReader reader, long iSize)
        {
            ListHeader lh = ReadListHeader(reader);
            if (!lh.bValid)
            {
                AddError("A LIST chunk was read, but its header was invalid.");
                return false;
            }

            long iPosition = lh.iPosition;

            lh.iDataSize = iSize;

            if (lh.iType == StringInt("hdrl"))
            {
                PosDiag("Header list", iPosition);
                m_lhHeaderArray.Add(lh);
                ParseMainHeader(reader, iSize);
            }
            else if (lh.iType == StringInt("movi"))
            {
                PosDiag("Movie list", iPosition);
                m_lhMovieArray.Add(lh);

                SkipChunk(reader, lh.iDataSize - 4);
            }
            else if (lh.iType == StringInt("strl"))
            {
                PosDiag("Stream list", iPosition);
                m_lhStreamArray.Add(lh);

                ParseStreamList(reader, iSize);
            }
            else if (lh.iType == StringInt("odml"))
            {
                PosDiag("OpenDML list", iPosition);
                m_lhStreamArray.Add(lh);

                ParseStreamList(reader, iSize);
            }
            else
            {
                PosDiag("(Unknown LIST)", iPosition);
                AddWarning(string.Format("An unknown LIST type (\"{0}\") was found at 0x{1:X16}", IntString(lh.iType), reader.BaseStream.Position - 8));
                SkipChunk(reader, lh.iDataSize - 4);
            }

            return true;
        }

        private void ParseStreamList(BinaryReader reader, long iSize)
        {
            long iMax = reader.BaseStream.Position + iSize;
            int iType;

            PushList();

            while (reader.BaseStream.Position < iMax)
            {
                long iPosition = reader.BaseStream.Position;

                iType = reader.ReadInt32();
                iSize = reader.ReadInt32();

                if (iType == StringInt("strh"))
                {
                    PosDiag("Stream header", iPosition);
                    ParseStreamHeader(reader, iSize);
                }
                else if (iType == StringInt("strf"))
                {
                    PosDiag("Stream format", iPosition);
                    ParseStreamFormat(reader, iSize);
                }
                else if (iType == StringInt("strd"))
                {
                    PosDiag("Stream data", iPosition);
                    SkipChunk(reader, iSize);
                }
                else if (iType == StringInt("strn"))
                {
                    PosDiag("Stream name", iPosition);
                    SkipChunk(reader, iSize);
                }
                else if (iType == StringInt("indx"))
                {
                    PosDiag("OpenDML index", iPosition);
                    ParseOpenDMLIndex(reader, iSize);
                }
                else if (iType == StringInt("dmlh"))
                {
                    PosDiag("OpenDML header", iPosition);
                    SkipChunk(reader, iSize);
                }
                else if (iType == StringInt("LIST"))
                {
                    ReadList(reader, iSize);
                }
                else if (iType == StringInt("JUNK"))
                {
                    PosDiag("JUNK chunk (ignored)", iPosition);
                    SkipChunk(reader, iSize);
                }
                else
                {
                    PosDiag("(unknown chunk)", iPosition);
                    AddWarning(string.Format("An unknown chunk type (\"{0}\") was found.", IntString(iType)));
                    SkipChunk(reader, iSize);
                }
            }

            PopList();

        }

        private void ParseMainHeader(BinaryReader reader, long iSize)
        {
            long iMax = reader.BaseStream.Position + iSize;
            int iType;

            while (reader.BaseStream.Position < iMax)
            {
                long iPosition = reader.BaseStream.Position;

                iType = reader.ReadInt32();
                iSize = reader.ReadInt32();

                if (iType == StringInt("avih"))
                {
                    PosDiag("AVI Header",iPosition);
                    ParseAVIHeader(reader, iSize);
                }
                else if (iType == StringInt("LIST"))
                {
                    ReadList(reader, iSize);
                }
                else if (iType == StringInt("JUNK"))
                {
                    PosDiag("JUNK chunk (ignored)", iPosition);
                    SkipChunk(reader, iSize);
                }
                else
                {
                    PosDiag("(unknown chunk)", reader.BaseStream.Position - 8);
                    AddWarning(string.Format("An unknown chunk type (\"{0}\") was found.", IntString(iType)));
                    SkipChunk(reader, iSize);
                }
            }
        }

        private void ParseStreamHeader(BinaryReader reader, long iSize)
        {
            long iMax = reader.BaseStream.Position + iSize;

            AVIStreamHeader hdr = null;

            int iType = reader.ReadInt32();
            m_iCurrentStreamHeaderType = iType;

            if (iType == StringInt("vids"))
            {
                hdr = m_ashVideo;
            }
            else if (iType == StringInt("auds"))
            {
                hdr = m_ashAudio;
            }

            if (hdr != null)
            {
                hdr.fccStreamType = iType;
                if (reader.BaseStream.Position < iMax)
                    hdr.fccStreamHandler = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    hdr.iFlags = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    hdr.iPriority = reader.ReadInt16();
                if (reader.BaseStream.Position < iMax)
                    hdr.iLanguage = reader.ReadInt16();
                if (reader.BaseStream.Position < iMax)
                    hdr.iInitialFrames = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    hdr.iSampleRateDenominator = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    hdr.iSampleRateNumerator = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    hdr.iStart = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    hdr.iLength = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    hdr.iSuggestedBufferSize = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    hdr.iQuality = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    hdr.iSampleSize = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    hdr.iLeft = reader.ReadInt16();
                if (reader.BaseStream.Position < iMax)
                    hdr.iTop = reader.ReadInt16();
                if (reader.BaseStream.Position < iMax)
                    hdr.iRight = reader.ReadInt16();
                if (reader.BaseStream.Position < iMax)
                    hdr.iBottom = reader.ReadInt16();
            }

            SkipChunk(reader, iMax - reader.BaseStream.Position);
        }

        private void ParseStreamFormat(BinaryReader reader, long iSize)
        {
            long iMax = reader.BaseStream.Position + iSize;
            int iType = m_iCurrentStreamHeaderType;

            if (iType == StringInt("vids"))
            {
                if (reader.BaseStream.Position < iMax)
                    m_asfVideo.iBMPSize = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    m_asfVideo.iBMPWidth = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    m_asfVideo.iBMPHeight = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    m_asfVideo.iBitPlanes = reader.ReadInt16();
                if (reader.BaseStream.Position < iMax)
                    m_asfVideo.iBitsPerPixel = reader.ReadInt16();
                if (reader.BaseStream.Position < iMax)
                    m_asfVideo.iCompression = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    m_asfVideo.iImageSize = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    m_asfVideo.iHorizPixelsPerMeter = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    m_asfVideo.iVertPixelsPerMeter = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    m_asfVideo.iColorsUsed = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    m_asfVideo.iColorsImportant = reader.ReadInt32();
            }
            else if (iType == StringInt("auds"))
            {
                if (reader.BaseStream.Position < iMax)
                    m_asfAudio.iIDTag = reader.ReadInt16();
                if (reader.BaseStream.Position < iMax)
                    m_asfAudio.iChannels = reader.ReadInt16();
                if (reader.BaseStream.Position < iMax)
                    m_asfAudio.iSamplesPerSecond = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    m_asfAudio.iAvgBytesPerSec = reader.ReadInt32();
                if (reader.BaseStream.Position < iMax)
                    m_asfAudio.iBlockAlign = reader.ReadInt16();
                if (reader.BaseStream.Position < iMax)
                    m_asfAudio.iBitsPerSample = reader.ReadInt16();
                if (reader.BaseStream.Position < iMax)
                    m_asfAudio.iFormatSpecificDataSize = reader.ReadInt16();
            }

            SkipChunk(reader, iMax - reader.BaseStream.Position);
        }

        private ListHeader ReadListHeader(BinaryReader reader)
        {
            ListHeader lh = new ListHeader();

            lh.iPosition = reader.BaseStream.Position - 8;
            lh.iType = reader.ReadInt32();

            lh.bValid = true;
            return lh;
        }

        private AVIIndex ReadAVIIndex(BinaryReader reader, long iSize)
        {
            AVIIndex ai = new AVIIndex();

            ai.iPosition = reader.BaseStream.Position;
            ai.iDataSize = iSize;

            Diag("Parsing AVI index");

            m_avi1Index = new ArrayList();

            long iMax = reader.BaseStream.Position + iSize;
            int iFrame = 0;

            while (reader.BaseStream.Position < iMax)
            {
                AVI1IndexEntry entry = new AVI1IndexEntry();

                entry.iChunkId = reader.ReadInt32();
                entry.iFlags = reader.ReadInt32();
                entry.iOffset = reader.ReadInt32();
                entry.iSize = reader.ReadInt32();

                bool bVideo = (((entry.iChunkId & 0x00ff0000) >> 16) == 'd');

                // See if this is a keyframe
                if ((entry.iFlags & Global.AVIIF_KEYFRAME) > 0)
                {
                    if (bVideo)
                    {
                        entry.iFrame = iFrame;
                        m_avi1Index.Add(entry);
                    }
                }

                if (bVideo)
                    iFrame++;
            }

            SkipChunk(reader, iMax - reader.BaseStream.Position);

            ai.bValid = true;
            return ai;
        }

        public long[] Keyframes
        {
            get
            {
                if (m_odmlFullIndex == null)
                {
                    if (m_KeyFrames == null)
                        m_KeyFrames = new long[m_avi1Index.Count];

                    if (m_KeyFrames.Length != m_avi1Index.Count)
                        m_KeyFrames = new long[m_avi1Index.Count];

                    for (int i = 0; i < m_avi1Index.Count; i++)
                    {
                        m_KeyFrames[i] = ((AVI1IndexEntry) m_avi1Index[i]).iFrame;
                    }
                    return m_KeyFrames;
                }
                else
                {
                    if (m_KeyFrames == null)
                        m_KeyFrames = new long[m_odmlFullIndex.Count];

                    if (m_KeyFrames.Length != m_odmlFullIndex.Count)
                        m_KeyFrames = new long[m_odmlFullIndex.Count];

                    for (int i = 0; i < m_odmlFullIndex.Count; i++)
                    {
                        m_KeyFrames[i] = ((OpenDMLIndexEntry) m_odmlFullIndex[i]).iFrame;
                    }
                    return m_KeyFrames;
                }
            }
        }

        private bool ReadRIFF(BinaryReader reader)
        {
            int i;

            i = reader.ReadInt32();
            if (i != StringInt("RIFF"))
            {
                AddError(string.Format("RIFF header not found (found \"{0}\" instead)", IntString(i)));
                return false;
            }

            m_iFileSize32bit = reader.ReadInt32();

            i = reader.ReadInt32();
            if (i != StringInt("AVI "))
            {
                AddError(string.Format("RIFF file is not an AVI (type \"{0}\")", IntString(i)));
                return false;
            }
            
            return true;
        }

        public string BasicInfo()
        {
            string sResult = "";

            sResult += string.Format("File Name:   {0}\n", Path.GetFileName(m_sAVIFile));
            sResult += string.Format("File Size:   {0} bytes\n", m_iFileSize);
            sResult += string.Format("FOURCC:      {0}\n", IntString(m_ashVideo.fccStreamHandler));
            sResult += string.Format("us/Frame:    {0}\n", m_aviBasicInfo.iMicroSecPerFrame);
            sResult += string.Format("Samplerate:  {0:F4}\n", m_ashVideo.iSampleRateNumerator / (double) m_ashVideo.iSampleRateDenominator);
            sResult += string.Format("Frames:      {0}\n", FrameCount);
            sResult += string.Format("Streams:     {0}\n", m_aviBasicInfo.iStreams);
            sResult += string.Format("Resolution:  {0}x{1}\n", m_aviBasicInfo.iWidth, m_aviBasicInfo.iHeight);
            sResult += string.Format("Compression: {0}\n", 
                m_ashVideo.fccStreamHandler > 0 ?
                  IntString(m_ashVideo.fccStreamHandler) :
                  "RAW");

            return sResult;
        }

        public long FrameCount
        {
            get
            {
                if (m_iMasterODMLFrameIndex > 0)
                    return m_iMasterODMLFrameIndex;

                return m_aviBasicInfo.iTotalFrames;
            }
        }

        public long LastKeyframe
        {
            get
            {
                return Keyframes[Keyframes.Length - 1];
            }
        }

        public string AdvancedInfo()
        {
            string sResult = "";

            foreach(ListHeader lh in m_lhHeaderArray)
            {
                sResult += string.Format("Header list, offset: 0x{0:X16}, size: 0x{0:X16}\n", lh.iPosition, lh.iDataSize);
            }
            foreach(ListHeader lh in m_lhStreamArray)
            {
                sResult += string.Format("Stream list, offset: 0x{0:X16}, size: 0x{0:X16}\n", lh.iPosition, lh.iDataSize);
            }
            foreach(ListHeader lh in m_lhMovieArray)
            {
                sResult += string.Format("Movie list,  offset: 0x{0:X16}, size: 0x{0:X16}\n", lh.iPosition, lh.iDataSize);
            }
            foreach(AVIIndex ai in m_aiArray)
            {
                sResult += string.Format("Index list,  offset: 0x{0:X16}, size: 0x{0:X16}\n", ai.iPosition, ai.iDataSize);
            }

            return sResult;
        }

        public void AddWarning(string s)
        {
            m_sWarnings += ("Warning: " + s + "\n");
        }

        public void AddError(string s)
        {
            m_sErrors += ("Error: " + s + "\n");
        }

        public static int StringInt(string s)
        {
            s += "\0\0\0\0";

            int i = 
                (
                    (s[3] << 24) |
                    (s[2] << 16) |
                    (s[1] << 8) |
                    s[0]
                );

            return i;
        }

        public static string IntString(int i)
        {
            return string.Format("{3}{2}{1}{0}",
                ((char) ((i >> 24) & 0xff)),
                ((char) ((i >> 16) & 0xff)),
                ((char) ((i >> 8) & 0xff)),
                ((char) (i & 0xff))
                );
        }

        public void ParseAVIHeader(BinaryReader reader, long iSize)
        {
            long iMax = reader.BaseStream.Position + iSize;

            if (reader.BaseStream.Position < iMax)
                m_aviBasicInfo.iMicroSecPerFrame = reader.ReadInt32();
            if (reader.BaseStream.Position < iMax)
                m_aviBasicInfo.iMaxBytesPerSec = reader.ReadInt32();
            if (reader.BaseStream.Position < iMax)
                m_aviBasicInfo.iPaddingGranularity = reader.ReadInt32();
            if (reader.BaseStream.Position < iMax)
                m_aviBasicInfo.iFlags = reader.ReadInt32();
            if (reader.BaseStream.Position < iMax)
                m_aviBasicInfo.iTotalFrames = reader.ReadInt32();
            if (reader.BaseStream.Position < iMax)
                m_aviBasicInfo.iInitialFrames = reader.ReadInt32();
            if (reader.BaseStream.Position < iMax)
                m_aviBasicInfo.iStreams = reader.ReadInt32();
            if (reader.BaseStream.Position < iMax)
                m_aviBasicInfo.iSuggestedBufferSize = reader.ReadInt32();
            if (reader.BaseStream.Position < iMax)
                m_aviBasicInfo.iWidth = reader.ReadInt32();
            if (reader.BaseStream.Position < iMax)
                m_aviBasicInfo.iHeight = reader.ReadInt32();

            SkipChunk(reader, iMax - reader.BaseStream.Position);
        }

        private void ReadOpenDMLFullIndex(OpenDMLIndex idx, BinaryReader reader)
        {
            m_odmlFullIndex = new ArrayList();
            m_iMasterODMLFrameIndex = 0;

            foreach(OpenDMLSuperIndexEntry entry in idx.arrayIndices)
            {
                reader.BaseStream.Position = entry.iOffset;

                int iFourCC = reader.ReadInt32();
                int iSize = reader.ReadInt32();
                ParseOpenDMLIndex(reader, iSize);
            }
        }

        private void ParseOpenDMLIndex(BinaryReader reader, long iSize)
        {
            long iMax = reader.BaseStream.Position + iSize;

            OpenDMLIndex idx = new OpenDMLIndex();

            if (reader.BaseStream.Position < iMax)
                idx.iLongsPerEntry = reader.ReadInt16();
            if (reader.BaseStream.Position < iMax)
                idx.bIndexSubType = reader.ReadByte();
            if (reader.BaseStream.Position < iMax)
                idx.bIndexType = reader.ReadByte();
            if (reader.BaseStream.Position < iMax)
                idx.iEntriesInUse = reader.ReadInt32();
            if (reader.BaseStream.Position < iMax)
                idx.iChunkID = reader.ReadInt32();
            if (reader.BaseStream.Position < iMax)
                idx.iBaseOffset = reader.ReadInt64();
            if (reader.BaseStream.Position < iMax)
                idx.iReserved = reader.ReadInt32();

            bool bVideo = (((idx.iChunkID & 0x00ff0000) >> 16) == 'd');
            bool bAudio = (((idx.iChunkID & 0x00ff0000) >> 16) == 'w');

            if (idx.bIndexType == Global.AVI_INDEX_OF_INDEXES)
            {
                OpenDMLSuperIndexEntry entry;

                int iIndex = 0;
                while (iIndex < idx.iEntriesInUse)
                {
                    entry = new OpenDMLSuperIndexEntry();

                    entry.iOffset = reader.ReadInt64();
                    entry.iSize = reader.ReadInt32();
                    entry.iDuration = reader.ReadInt32();

                    idx.arrayIndices.Add(entry);
                    iIndex++;

                    if (reader.BaseStream.Position >= iMax)
                        break;
                }

                if (bVideo)
                {
                    m_odmlIndexVideo = idx;
                }
                else if (bAudio)
                {
                    m_odmlIndexAudio = idx;
                }

            }
            else if (idx.bIndexType == Global.AVI_INDEX_OF_CHUNKS)
            {
                OpenDMLIndexEntry entry;

                int iIndex = 0;
                while (iIndex < idx.iEntriesInUse)
                {
                    entry = new OpenDMLIndexEntry();
                    bool bKeyframe = true;

                    entry.iOffset = reader.ReadInt32() + idx.iBaseOffset;
                    entry.iSize = reader.ReadInt32();

                    if ((entry.iSize & 0x80000000) > 0)
                        bKeyframe = false;

                    if (idx.bIndexSubType == Global.AVI_INDEX_2FIELD)
                    {
                        entry.iOffsetField2 = reader.ReadInt32() + idx.iBaseOffset;
                        entry.bFieldIndex = true;
                    }

                    entry.iFrame = m_iMasterODMLFrameIndex;

                    if (bKeyframe)
                        m_odmlFullIndex.Add(entry);

                    iIndex++;
                    m_iMasterODMLFrameIndex++;

                    if (reader.BaseStream.Position >= iMax)
                        break;
                }
            }

            SkipChunk(reader, iMax - reader.BaseStream.Position);
        }

    }

    public class ListHeader
    {
        public long iPosition;
        public long iDataSize;
        public long iDataOffset;
        public int iType;
        public bool bValid;

        public ListHeader()
        {
            iPosition = -1;
            iDataSize = -1;
            iType = -1;
            bValid = false;
            iDataOffset = 8;
        }

        public ListHeader(ListHeader copy)
        {
            iPosition = copy.iPosition;
            iDataSize = copy.iDataSize;
            iDataOffset = copy.iDataOffset;
            iType = copy.iType;
            bValid = copy.bValid;
        }
    }

    public class AVIIndex
    {
        public long iPosition;
        public long iDataSize;
        public long iDataOffset;
        public int iType;
        public bool bValid;

        public AVIIndex()
        {
            iPosition = -1;
            iDataSize = -1;
            iType = -1;
            bValid = false;
            iDataOffset = 8;
        }
    }

    public class OpenDMLIndex
    {
        public short iLongsPerEntry;
        public byte bIndexSubType;
        public byte bIndexType;
        public int iEntriesInUse;
        public int iChunkID;
        public long iBaseOffset;
        public int iReserved;

        public ArrayList arrayIndices;

        public OpenDMLIndex()
        {
            iLongsPerEntry = 0;
            bIndexSubType = 0;
            bIndexType = 0;
            iEntriesInUse = 0;
            iChunkID = 0;
            iBaseOffset = 0;
            iReserved = 0;

            arrayIndices = new ArrayList();
        }
    }

    public class OpenDMLSuperIndexEntry
    {
        public long iOffset;
        public int iSize;
        public int iDuration;

        public OpenDMLSuperIndexEntry()
        {
            iOffset = 0;
            iSize = 0;
            iDuration = 0;
        }
    }

    public class OpenDMLIndexEntry
    {
        public long iOffset;
        public int iSize;
        public long iOffsetField2;
        public bool bFieldIndex;
        public long iFrame;

        public OpenDMLIndexEntry()
        {
            iOffset = 0;
            iSize = 0;
            iOffsetField2 = 0;
            bFieldIndex = false;
            iFrame = 0;
        }
    }

    public class AVIBasicInfo
    {
        public int iFourCC;
        public int iSize;
        public int iMicroSecPerFrame;
        public int iMaxBytesPerSec;
        public int iPaddingGranularity;
        public int iFlags;
        public int iTotalFrames;
        public int iInitialFrames;
        public int iStreams;
        public int iSuggestedBufferSize;
        public int iWidth;
        public int iHeight;

        public AVIBasicInfo()
        {
            iFourCC = 0;
            iSize = 0;
            iMicroSecPerFrame = 0;
            iMaxBytesPerSec = 0;
            iPaddingGranularity = 0;
            iFlags = 0;
            iTotalFrames = 0;
            iInitialFrames = 0;
            iStreams = 0;
            iSuggestedBufferSize = 0;
            iWidth = 0;
            iHeight = 0;
        }
    }

    public class AVIStreamHeader
    {
        public int fccStreamType;
        public int fccStreamHandler;
        public int iFlags;
        public short iPriority;
        public short iLanguage;
        public int iInitialFrames;
        public int iSampleRateDenominator;
        public int iSampleRateNumerator;
        public int iStart;  // time (in samples) to start first stored sample
        public int iLength; // Stream length in samples
        public int iSuggestedBufferSize;
        public int iQuality;
        public int iSampleSize;
        public short iLeft;
        public short iTop;
        public short iRight;
        public short iBottom;

        public AVIStreamHeader()
        {
            fccStreamType = 0;
            fccStreamHandler = 0;
            iFlags = 0;
            iPriority = 0;
            iLanguage = 0;
            iInitialFrames = 0;
            iSampleRateDenominator = 0;
            iSampleRateNumerator = 0;
            iStart = 0;
            iLength = 0;
            iSuggestedBufferSize = 0;
            iQuality = 0;
            iSampleSize = 0;
            iLeft = 0;
            iTop = 0;
            iRight = 0;
            iBottom = 0;
        }
    }

    public class AVIStreamFormatVideo
    {
        public int iBMPSize;
        public int iBMPWidth;
        public int iBMPHeight;
        public short iBitPlanes;
        public short iBitsPerPixel;
        public int iCompression;        // e.g. "DX50" - zero is uncompressed
        public int iImageSize;          // uncompressed image size
        public int iHorizPixelsPerMeter;
        public int iVertPixelsPerMeter;
        public int iColorsUsed;
        public int iColorsImportant;

        public AVIStreamFormatVideo()
        {
            iBMPSize = 0;
            iBMPWidth = 0;
            iBMPHeight = 0;
            iBitPlanes = 0;
            iBitsPerPixel = 0;
            iCompression = 0;
            iImageSize = 0;
            iHorizPixelsPerMeter = 0;
            iVertPixelsPerMeter = 0;
            iColorsUsed = 0;
            iColorsImportant = 0;
        }
    }

    public class AVIStreamFormatAudio
    {
        public short iIDTag;
        public short iChannels;
        public int iSamplesPerSecond;
        public int iAvgBytesPerSec;
        public short iBlockAlign;
        public short iBitsPerSample;
        public short iFormatSpecificDataSize;

        public AVIStreamFormatAudio()
        {
            iIDTag = 0;
            iChannels = 0;
            iSamplesPerSecond = 0;
            iAvgBytesPerSec = 0;
            iBlockAlign = 0;
            iBitsPerSample = 0;
            iFormatSpecificDataSize = 0;
        }
    }

    public class AVI1IndexEntry
    {
        public int iChunkId;
        public int iFlags;
        public int iOffset;
        public int iSize;
        public int iFrame;

        public AVI1IndexEntry()
        {
            iChunkId = 0;
            iFlags = 0;
            iOffset = 0;
            iSize = 0;
            iFrame = 0;
        }
    }
}
